#!/usr/bin/perl -w
##############################################################################
# Analyzes data about application collected by apptodb.pl.
#
# Author: Denis Silakov <silakov@ispras.ru>
##############################################################################

use strict;
use Getopt::Long;
use Text::Wrap;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

# Options
my $opt_help;

my $FileName;
my $IntFile;
my $LibFile;
my $RequiredLibsFile;
my $RequiredIntsFile;
my $outdir = ".";

my $lsbversion;
my $arch;

# Interfaces provided by application ELF files
# Key: Interface name (unmangled, if any)
# Value: list of commands or shared objects (file names) that provide the interface.
my %shared_ints;

# Interfaces required by application executables, organized as 'shared_ints'.
# Interfaces included in this hash and not included in 'shared_ints'
# are external interfaces that the application wants to get
# from the system libraries
my %linked_ints;

# Application's ELF files may require some shared libraries.
# Let's check, which of these libraries are provided by the application itself
# and which should be provided by system
my %needed_libs;
my %provided_libs;

# Information about interfaces and libraries included in the LSB
# lsb_libs: key:   library runname
#           value: =Lname (doesn't matter, if a value for a library 'key' is defined, then this library is included in the LSB
# lsb_ints: key:   interface name (mangled, if any)
#           value: hash; key - Ilibrary, value = 1 (the same as for lsb_libs)
my %lsb_libs;
my %lsb_ints;
my %lsb_lib_runnames;

#  key:   (Iname,Ilibrary)
#  value: Interface version name
my %lsb_int_versions;

# Origins
my %lib_origins;
my %int_origins;

my %unmangled;

sub HELP_MESSAGE();

# Process args
GetOptions("help|h!" => \$opt_help,
			"outdir|d=s" => \$outdir);

HELP_MESSAGE() if (defined $opt_help);
if( @ARGV != 3 ) {
	HELP_MESSAGE();
	exit(1);
}

$RequiredLibsFile = $outdir."/required_libs";
$RequiredIntsFile = $outdir."/required_ints";

$lsbversion = $ARGV[0];
$arch = $ARGV[1];
$FileName = $ARGV[2];

my $DOCUMENT_ROOT = $0;
$DOCUMENT_ROOT =~ s/\/[^\/]+$//;

$IntFile = "$DOCUMENT_ROOT/Int_".$lsbversion."_".$arch;
$LibFile = "$DOCUMENT_ROOT/Lib_".$lsbversion."_".$arch;

# Check if all files exist
if( !-e $FileName ) {
	die "Can't open data file $FileName";
}
if( !-e $IntFile ) {
	die "Can't open file with LSB interfaces ($IntFile)";
}
if( !-e $LibFile ) {
	die "Can't open file with LSB libraries ($LibFile)";
}

##############################################################################
# Main section
##############################################################################

my %libhash;

open( INT_DATA, "<$IntFile" );
while( <INT_DATA> ) {
	# correct interface record should contain Iname, Ilibrary, Iversion, Iunmangled
	if( /'([^']+)';'([^']*)';'([^']*)';'([^']*)'/ ) {
		if( defined $lsb_ints{$1} ) {
			%libhash = %{$lsb_ints{$1}};
		}
		else {
			%libhash = ();
		}
 		$libhash{$2} = 1;
		$lsb_ints{$1} = {%libhash};
		$lsb_int_versions{($1,$2)} = $3;
		$unmangled{$1} = $4;
	}
	else {
		print STDERR "Incorrrect lsb interface string: $_\n";
	}
}
close( INT_DATA );

open( LIB_DATA, "<$LibFile" );
while( <LIB_DATA> ) {
	# correct library record should contain Lname, ALrunname
	if( /'([^']+)';'([^']+)'/ ) {
		$lsb_libs{$2} = $1;
		$lsb_lib_runnames{$1} = $2;
	}
}
close( LIB_DATA );

# foreach my $intname (keys %lsb_ints) {
# 	%libhash = %{$lsb_ints{$intname}};
# 	foreach my $libname (keys %libhash) {
# 		print "Int: '".$intname."';'".$libname;
# 		print "';'".$lsb_int_versions{($intname,$libname)}."';'".$unmangled{$intname}."'\n";
# 	}
# }
#
# exit(0);

open( UPLOAD_DATA, "<$FileName" );

my $symtab=0;    # indicates that we are processing 'symtab' section
my $dynsym=0;    # indicates that we are processing 'dynsym' section
my $cmd=0;       # indicates that we are processing output of a command
my $lib=0;       # indicates that we are processing output of a shared object
my $dynsec=0;    # indicates that we are processing general information about dynamic section
my $elfheader=0; # indicates that we are processing general information about dynamic section
my $skip=0;      # indicates that we are skipping lines until 'Library' or 'Command' record

my $origin_name="";
my @origin_list;

while( <UPLOAD_DATA> ) {
	if( ($dynsec or $dynsym) and ( /^\S/ ) ) {
		$dynsec=0;
		$dynsym=0;
	}

	if( /^ELF Header/ ) {
		$elfheader=1;
	}
	elsif ( /^\S/ ) {
		$elfheader=0;
	}

	if( $elfheader and /Class\:\s*ELF(\d+)/ ) {
		$skip = 0;
		my $elf_arch = $1;

		if( $elf_arch eq "32" ) {
			if( $arch =~ /x86-64/i or $arch =~ /IA64/i or $arch =~ /S390X/i or $arch =~ /PPC64/i ) {
				$skip = 1;
			}
		}
		elsif( $elf_arch eq "64" ) {
			if( $arch =~ /PPC32/i or $arch =~ /IA32/i or $arch =~ /S390/i ) {
				$skip = 1;
			}
		}
		else {
			print STDERR "Strange 'Class' in ELF Header - $elf_arch\n";
		}
	}

	if( /^\!Component/ ) {
		# new component, reset flags
		$symtab=0;
		$dynsym=0;
		$dynsec=0;
		$cmd=0;
		$lib=0;
		$skip=0;
	}
	elsif( /^\!Library/ ) {
		# new shared object, reset flags
		$symtab=0;
		$dynsym=0;
		$dynsec=0;
		$cmd=0;
		$lib=1;
		$skip=0;

		# The string should contain name, version, runname, path, ABI tag (optional) and so-name (optional) of the library
		if( /\!Library '(([^\'])+)' '(([^\'])*)' '(([^\'])+)' '(([^\'])+)' '(([^\'])*)' '(([^\'])*)'/ ) {
			my $runname = $5;
			my $soname = $11;

			$origin_name = "$7/$runname";
			$provided_libs{$runname} = 1;
			if( $soname ) {
				$provided_libs{$soname} = 1;
			}

			if( $runname =~ /(.+\.so)\./ ) {
				$provided_libs{$1} = 1;
			}
		}
		else {
			print STDERR "Incorrect library record at line $.:\n\t'$_'\n";
		}
	}
	elsif( /^Dynamic\ section/ ) {
		if( $skip ) {
			next;
		}
		$symtab=0;
		$dynsym=0;
		$dynsec=1;
	}
	elsif( /'.dynsym'/ ) {
		if( $skip ) {
			next;
		}
		$symtab=0;
		$dynsym=1;
		$dynsec=0;
	}
	elsif( /'.symtab'/ ) {
		if( $skip ) {
			next;
		}
		$symtab=1;
		$dynsym=0;
		$dynsec=0;
	}
	elsif( /NEEDED\)\s+Shared\s+library\:\s+\[((\w|-|.)+)\]/ ) {
		if( $skip ) {
			next;
		}

		if( $dynsec == 1 ) {
			my $fullname = $1;
			my $libname;

			if( $fullname =~ /\// ) {
				my @libname = split(/\//,$fullname);
				my $realname = $libname[$#libname];
				if( $realname eq "/" ) {
				    $realname = $libname[$#libname-1];
				}
				$needed_libs{$realname} = 1;
				$libname = $realname;
			}
			else {
				$needed_libs{$fullname} = 1;
				$libname = $fullname;
			}

			if( defined $lib_origins{$libname} ) {
				@origin_list = @{$lib_origins{$libname}};
			}
			else {
				@origin_list = ();
			}
			push @origin_list, $origin_name;
			$lib_origins{$libname} = [ @origin_list ];
		}
	}
	elsif( /\s*\d+:\s+(\w*)\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/ ) {
		# (the line of 'readelf' output corresponding to the interface)

		if( $skip ) {
			next;
		}

		# we also need to know if the interface is in the shared object
		# or in the library
		if( ($lib==0) and ($cmd==0) ) {
			print STDERR "Cannot determine if interface $1 is in executable or in shared object!\n";
			exit(1);
		}

		# do nothing with symtab (but there are some plans for the future)
		if( ($symtab == 1) or ($dynsym == 0) ) {
			next;
		}

		my $fullname=$6; # name, maybe with version
		my $idx=$1;      # offset
 		my $type=$2;     # FUNC, OBJECT etc.
 		my $bind=$3;     # GLOBAL, WEAK or LOCAL
 		my $vis=$4;      # Visibility
		my $Ndx=$5;		 #

		# Filter interfaces by their binding, type and visibility
		if( ($bind ne 'WEAK') and ($bind ne 'GLOBAL') ) {
			next;
		}
		if( ($type ne 'FUNC') and ($type ne 'OBJECT') and ($type ne 'COMMON') ) {
			next;
		}
		if( $vis ne 'DEFAULT' ) {
			next;
		}

		# If we are processing executable, ignore internal interfaces
		# otherwise ignore exported interfaces
#		if( ($cmd==1) and ($idx =~ /\D|1|2|3|4|5|6|7|8|9/) ) {
		if( ($cmd==1) and ($Ndx ne "UND") ) {
			if( $fullname =~ /\@\@/ ) {
				my $fullname_single = $fullname;
				$fullname_single =~ s/\@\@/\@/;
				$shared_ints{$fullname_single} = 1;
				my ($short_name,$others) = split( "@", $fullname_single );
				$shared_ints{$short_name} = 1;
			}
			elsif( $fullname =~ /\@/ ) {
				my $fullname_doubled = $fullname;
				$fullname_doubled =~ s/\@/\@\@/;
				$shared_ints{$fullname_doubled} = 1;
				my ($short_name,$others) = split( "@", $fullname );
				$shared_ints{$short_name} = 1;
			}
			$shared_ints{$fullname} = 1;

			next;
		}
#		if( ($lib==1) and ($idx !~ /\D|1|2|3|4|5|6|7|8|9/) ) {
		if( ($lib==1) and ($Ndx eq "UND") ) {
			if( defined $int_origins{$fullname} ) {
				@origin_list = @{$int_origins{$fullname}};
			}
			else {
				@origin_list = ();
			}
			push @origin_list, $origin_name;
			$int_origins{$fullname} = [ @origin_list ];

			$linked_ints{$fullname} = 1;
			next;
		}

		if( $cmd==1 ) {
			if( defined $int_origins{$fullname} ) {
				@origin_list = @{$int_origins{$fullname}};
			}
			else {
				@origin_list = ();
			}
			push @origin_list, $origin_name;
			$int_origins{$fullname} = [ @origin_list ];

			$linked_ints{$fullname} = 1;
			next;
		}
		if( $lib==1 ) {
			if( $fullname =~ /\@\@/ ) {
				my $fullname_single = $fullname;
				$fullname_single =~ s/\@\@/\@/;
				$shared_ints{$fullname_single} = 1;
				my ($short_name,$others) = split( "@", $fullname_single );
				$shared_ints{$short_name} = 1;
			}
			elsif( $fullname =~ /\@/ ) {
				my $fullname_doubled = $fullname;
				$fullname_doubled =~ s/\@/\@\@/;
				$shared_ints{$fullname_doubled} = 1;
				my ($short_name,$others) = split( "@", $fullname );
				$shared_ints{$short_name} = 1;
			}
			$shared_ints{$fullname} = 1;
		}
	}
	elsif( /^\!Command/ ) {
		# The string should contain name and path
		if( /\!Command '([^']+)' '([^']*)'/ ) {
			$origin_name = "$2";
		}
		else {
			print STDERR "Incorrect command record at line $.:\n\t'$_'\n";
		}

		# new command, reset flags
		$symtab=0;
		$dynsym=0;
		$dynsec=0;
		$cmd=1;
		$lib=0;
		$skip=0;
	}
}

close( UPLOAD_DATA );

# Proceed with collected libraries
open( LIBS, "|sort >$RequiredLibsFile" );
foreach my $runname (keys %needed_libs) {
	if( $provided_libs{$runname} ) {
		# Library '$runname' is provided by application, skipping
		next;
	}

	# For other libs, print a line containing the following values, semicolon separated:
	#     runname
	#     lsb status (1/0)
	#     list of origins, coma separated
	#     comment (not used at the moment)

	print LIBS "$runname;";

	if( $lsb_libs{$runname} ) {
		print LIBS "1;";
	}
	else {
		print LIBS "0;";
	}

	my @origins_list = @{$lib_origins{$runname}};
	my $origins_libs = join( ',', @origins_list );
	print LIBS $origins_libs.";\n";
}
close( LIBS );

# Now it's time to process all collected interfaces
my %processed_ints; # Track single interface names, without versions

open( INTS, "| sort >$RequiredIntsFile" );
foreach my $fullname (keys %linked_ints) {
	if( $shared_ints{$fullname} ) {
		# Interface '$fullname' is linked from application shared objects, skipping
		next;
	}

	# For other ints, print a line containing the following values, semicolon separated:
	#     name (mangled, without version)
	#     unmangled name (if any)
	#     version
	#     lsb status (1/0)
	#     list of origins, coma separated
	#     library
	#     comment (not used at the moment)

	(my $realname, my $version) = split /\@\@/, $fullname;
	if( not $version ) {
		($realname, $version) = split /\@/, $fullname;
	}

	chomp($realname);
	if($version) {
		chomp($version);
	}
	else {
		$version = '';
	}

	if( $processed_ints{$realname} ) {
		next;
	}

	my $unmangled = `c++filt $realname`;
	chomp($unmangled);
	if( $unmangled eq $realname ) {
		$unmangled = "";
	}

	print INTS "$realname;$unmangled;$version;";

	my %found_libs = ();
	my @origins_int_list = @{$int_origins{$fullname}};
	my $LibName = "";

	if( $lsb_ints{$realname} ) {
		%libhash = %{$lsb_ints{$realname}};
		foreach my $libname (keys %libhash) {
			my $runname = $lsb_lib_runnames{$libname};

			if( $lib_origins{$runname} ) {
				my @origins_lib_list = @{$lib_origins{$runname}};
				my $origins_libs = join( ',', @origins_lib_list );

				for( my $i=0; $i<@origins_int_list; $i++ ) {
					my $intname = $origins_int_list[$i];
					if( $origins_libs =~ /\Q$intname\E/ ) {
						$found_libs{$intname} = $libname;
						$LibName = $libname;
					}
				}
			}
		}

		my @correct_ints = keys %found_libs;
		if( $#origins_int_list == $#correct_ints ) {
			print INTS "1;";
		}
		else {
			if( $#origins_int_list > 0 ) {
				my $correct_ints_string = join( ',', @correct_ints );
				print INTS "1;";
				print INTS "$correct_ints_string;$LibName;\n";
				print INTS "$realname;$version;";
			}
			print INTS "0;";

			%libhash = %{$lsb_ints{$realname}};
			my @int_lsb_libs = (keys %libhash);
			for( my $i=0; $i<@int_lsb_libs; $i++ ) {
				$int_lsb_libs[$i] = "'".$lsb_lib_runnames{$int_lsb_libs[$i]}."'";
			}

			my $comment = "Interface '$realname' is included in the LSB and assigned to ";
			if( $#int_lsb_libs > 1 ) {
				my $int_lsb_libs_string = join( " and ", @int_lsb_libs );
				$comment.= $int_lsb_libs_string." libraries which are absent in the NEEDED list";
			}
			else {
				$comment.= $int_lsb_libs[0]." library which is absent in the NEEDED list";
			}

			# Do not mark interface as processed - other files can use interface with the same
			# name but with the different version, and that files can contain proper lib in their 'NEEDED' sections
# 			$processed_ints{$realname} = 1;

			my @incorrect_ints = ();
			for( my $i=0; $i<@origins_int_list; $i++ ) {
				if( !$found_libs{$origins_int_list[$i]} ) {
					push @incorrect_ints, $origins_int_list[$i];
				}
			}

			my $incorrect_ints_string = join( ',', @incorrect_ints );
 			print INTS $incorrect_ints_string.";;$comment\n";

			next;
		}

	}
	else {
		print INTS "0;";
	}

	$processed_ints{$realname} = 1;

	my $origins_ints = join( ',', @origins_int_list );
	print INTS $origins_ints.";$LibName;\n";

}
close( INTS );

##############################################################################
# Miscellaneous functions
##############################################################################

sub HELP_MESSAGE() {
	print STDERR <<"EOM"

    $0: Analyzes output collected by apptodb.pl. Creates two files, 'required_ints' and 'required_libs',
        containing interfaces and libraries that should be provided by system in order to run the application.

    Usage: $0 [options] lsb_version arch data_file

    lsb_version             LSB version
    arch                    Architecture. Possible values are IA32, IA64, PPC32,
                            PPC64, S390, S390X and x86-64.
    data_file               File with data to be processed
    -d|--outdir             Directory where output files will be placed.
    -h|--help               Display this help message

EOM
	; exit(1);
}

